import datetime
import logging
import time
import typing as T

import tinkoff.invest as ti

from market_database.common import utils
from market_database.seeker import utils as seeker_utils


logger = utils.get_logger(__name__)


class BatchingPeriod:
    BATCHING_MONTH = "month"
    BATCHING_3_MONTH = "3 month"
    BATCHING_6_MONTH = "6 month"


BATCHING_PERIOD_DELTA = {
    BatchingPeriod.BATCHING_MONTH: datetime.timedelta(days=30),
    BatchingPeriod.BATCHING_3_MONTH: datetime.timedelta(days=90),
    BatchingPeriod.BATCHING_6_MONTH: datetime.timedelta(days=180),
}


class HistorySeeker:
    def __init__(self, api_token: str, batching_period: str = BatchingPeriod.BATCHING_MONTH):
        self._api_token = api_token
        self._batching_period = batching_period

    def seek(
        self,
        ticker: str,
        from_date: datetime.date,
        to_date: datetime.date,
        interval_name: str,
    ) -> T.Set[ti.schemas.HistoricCandle]:
        figi = seeker_utils.get_ticker_figi(ticker)
        interval = utils.get_interval_from_name(interval_name)

        logger.info(f"Use batching period {self._batching_period}")

        candle_data = set()
        current_from_date = from_date
        previous_from_date = None
        current_to_date = None
        with ti.Client(self._api_token) as client:
            while previous_from_date != current_from_date or current_to_date != to_date:
                previous_from_date = current_from_date
                current_to_date = min(current_from_date + BATCHING_PERIOD_DELTA[self._batching_period], to_date)

                logger.info(f"Current request interval: {current_from_date} - {current_to_date}")
                try:
                    current_candles = list(
                        client.get_all_candles(figi=figi, from_=current_from_date, to=current_to_date, interval=interval)
                    )
                    candle_data.update(current_candles)

                    current_from_date = max(candle.time for candle in current_candles).replace(tzinfo=None)
                except ti.exceptions.RequestError as e:
                    sleep_time = e.metadata.ratelimit_reset
                    logger.info(f"Wait {sleep_time} seconds")
                    time.sleep(sleep_time)

        return candle_data



from market_database.common import utils


def get_ticker_figi(ticker: str) -> str:
    figi_map = utils.get_ticker_figi_map()
    if ticker not in figi_map:
        raise NotImplementedError(f"FIGI mapping for ticker {ticker} is not implemented now")

    return figi_map[ticker]



import datetime
from importlib import resources
import logging
from pathlib import Path
import pickle
import sys
import typing as T
import json

from tinkoff.invest import CandleInterval

from market_database.common.types import Resource


INTERVALS = {
    "1m": CandleInterval.CANDLE_INTERVAL_1_MIN,
    "1h": CandleInterval.CANDLE_INTERVAL_HOUR,
    "1d": CandleInterval.CANDLE_INTERVAL_DAY,
}
MAX_CANDLES_TIMEDELTA = {
    "1m": datetime.timedelta(minutes=1),
    "1h": datetime.timedelta(hours=1),
    "1d": datetime.timedelta(days=1),
}
TICKER_FIGI_MAP_RESOURCE = Resource("market_database.configs", "ticker_figi_map.json")


def read_text_resource(resource: Resource) -> str:
    return resources.read_text(resource.package, resource.name)


def read_json(path: Path) -> T.Dict:
    with open(path, 'r') as input_file:
        return json.load(input_file)
    

def read_pickle(path: Path) -> T.Any:
    with open(path, "rb") as input_file:
        return pickle.load(input_file)
    

def dump_pickle(values: T.Any, path: Path):
    with open(path, "wb") as output_file:
        pickle.dump(values, output_file)


def get_ticker_figi_map() -> T.Dict[str, str]:
    return json.loads(read_text_resource(TICKER_FIGI_MAP_RESOURCE))

def is_valid_interval_name(interval_name: str) -> bool:
    return interval_name in INTERVALS


def is_valid_ticker(ticker: str) -> bool:
    return ticker in get_ticker_figi_map()


def get_interval_from_name(interval_name: str) -> CandleInterval:
    if interval_name not in INTERVALS:
        raise NotImplementedError(f"Interval with name {interval_name} is not implemented")
    
    return INTERVALS[interval_name]


def get_max_candles_delta(interval_name: str) -> datetime.timedelta:
    if interval_name not in INTERVALS:
        raise NotImplementedError(f"Interval with name {interval_name} is not implemented")
    
    return MAX_CANDLES_TIMEDELTA[interval_name]


def get_logger(name: str) -> logging.Logger:
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)

    return logger



import datetime
from pathlib import Path
import typing as T

from market_database.common import utils

from tinkoff.invest import HistoricCandle


class DataKeeper:
    def __init__(self, db_path: Path):
        self.db_path = db_path

    def _check_ticker_dir(self, ticker_dir: Path):
        for interval_dir in ticker_dir.iterdir():
            if not utils.is_valid_interval_name(interval_dir.name):
                raise RuntimeError(f"{interval_dir.name} is not a valid interval name with path {interval_dir}")

    def _check_database(self):
        for ticker_dir in self.db_path.iterdir():
            if not ticker_dir.is_dir():
                raise RuntimeError(f"Path {ticker_dir} is not a directory")
            if not utils.is_valid_ticker(ticker_dir.name):
                raise RuntimeError(f"{ticker_dir.name} is not a valid ticker with path {ticker_dir}")

    def _check_ticker(self, ticker: str):
        if not utils.is_valid_ticker(ticker):
            raise ValueError(f"{ticker} is not a valid ticker")

    def _check_interval_name(self, interval: str):
        if not utils.is_valid_interval_name(interval):
            raise ValueError(f"{interval} is not a valid interval name")

    def create_database(self):
        self.db_path.mkdir()

    def init(self):
        if not self.db_path.exists():
            self.create_database()
        self._check_database()

    def _get_ticker_data_root_path(self, ticker: str, interval: str) -> Path:
        return self.db_path / ticker / interval

    def _get_ticker_data_path(self, ticker: str, interval: str, date: datetime.date) -> Path:
        return self._get_ticker_data_root_path(ticker, interval) / date.isoformat()

    def _read_ticker_data(self, ticker: str, interval: str, date: datetime.date) -> T.Set[HistoricCandle]:
        ticker_data_path = self._get_ticker_data_path(ticker, interval, date)
        if not ticker_data_path.exists():
            return set()

        return utils.read_pickle(ticker_data_path)

    def _write_ticker_data(self, data: T.Set[HistoricCandle], ticker: str, interval: str, date: datetime.date):
        ticker_data_path = self._get_ticker_data_path(ticker, interval, date)
        utils.dump_pickle(data, ticker_data_path)

    def get_ticker_data(
        self, ticker: str, from_date: datetime.datetime, to_date: datetime.datetime, interval: str
    ) -> T.Set[HistoricCandle]:
        self._check_ticker(ticker)
        self._check_interval_name(interval)

        ticker_data = self._read_ticker_data(ticker, interval)
        return [candle for candle in ticker_data if candle.time >= from_date and candle.time < to_date]

    def set_ticker_data(self, data: T.Set[HistoricCandle], ticker: str, interval: str):
        self._check_ticker(ticker)
        self._check_interval_name(interval)

        ticker_data_root_path = self._get_ticker_data_root_path(ticker, interval)
        if not ticker_data_root_path.exists():
            ticker_data_root_path.mkdir(parents=True)

        unique_dates = {candle.time.date() for candle in data}
        for date in unique_dates:
            filtered_data = {candle for candle in data if candle.time.date == date}

            ticker_data = self._read_ticker_data(ticker, interval, date)
            ticker_data.update(filtered_data)

            self._write_ticker_data(ticker_data, ticker, interval, date)
